import uuid
import os


class TempFile:
    def __init__(self, base_path: str = "./", ext: str = ".jpg"):
        self.path = "{base_path}/{rand}{ext}".format(base_path=base_path,
                                                     rand=str(uuid.uuid4()),
                                                     ext=ext)

    def cleanup(self):
        os.remove(self.path)



from pathlib import Path
from typing import List, Any, Union
import json
import os
import csv
import pickle
import datetime
import shutil
import uuid

INFO = "INFO: "
WARNING = "WARNING: "
ERROR = "ERROR: "


def has_suffix(filename: Union[Path, str], suffixes: List[str]) -> bool:
    """Assumes that the filename path givens has a structure
    of directories separated with '/'. The last name is assumed
    to be the filename that we look to verify the suffix

   Parameters
   ----------
   filename
   suffixes

   Returns
   -------

   """

    if isinstance(filename, Path):
        filename_copy = str(filename)
    else:
        filename_copy = filename

    filename, file_extension = os.path.splitext(filename_copy)

    if file_extension in suffixes:
        return True

    return False





def read_json(filename: Path) -> dict:
    """Read the specified JSON file into
    a directory

    Parameters
    ----------
    filename: The path to the JSON file

    Returns
    -------

    An instance of dict with the properties
    described in the JSON file
    """

    with open(filename) as json_file:
        json_input = json.load(json_file)
        return json_input


def save_as_json(obj: Any, filename: Path) -> None:
    """Save the given object file in json

    Parameters
    ----------
    obj: The object to serialize
    filename: The path to the file

    Returns
    -------
    None
    """
    with open(filename, "w") as fp:
        json.dump(obj, fp)


def save_as_pickle(obj: Any, filename: Path,
                   protocol=pickle.HIGHEST_PROTOCOL) -> None:
    """Save the object in a pickle representation

    Parameters
    ----------
    obj: The object to save
    filename: The filename to use
    protocol: pickle protocol

    Returns
    -------

    None
    """

    with open(filename, 'wb') as outp:
        pickle.dump(obj, outp, protocol)


def create_dir_path(path: Path):
    if not os.path.exists(path=str(path)):
        os.makedirs(str(path))
    else:
        raise FileExistsError(f"File={path} already exists")


def create_dir(path: Path, dirname: str):
    """Create a directory with the given name in the specified path
    Throws a ValueError exception if the directory already exists

    Parameters
    ----------
    path: The path to create the directory
    dirname: The directory name

    Returns
    -------

    """

    dir_path = str(path) + "/" + dirname
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    else:
        raise ValueError("Could not create directory. The directory={0} exist".format(dir_path))


def save_list_as_csv(list_inst: List[Any], filename: Path, write_default_header: bool = False,
                     header: Any = None) -> None:
    """Save the given list in a csv file format

    Parameters
    ----------
    list_inst: The list to save
    filename: The filename to save at
    write_default_header: If true writes the default header which is simply
    the system date and time. In this case the row starts with a '#' mark
    header: Application defined header.
    Returns
    -------

    """
    with open(filename, "w") as fh:
        csv_writer = csv.writer(fh, delimiter=",")

        if header is not None:
            csv_writer.writerow(header)
        elif write_default_header:
            now = datetime.datetime.now()
            csv_writer.writerow("# " + str(now))

        for i, item in enumerate(list_inst):
            row = [i, item]
            csv_writer.writerow(row)


def load_list_from_csv(filename: Path) -> List[float]:
    """Loads a list form the given csv file

    Parameters
    ----------
    filename

    Returns
    -------

    A list of floats
    """
    with open(filename, 'r') as fh:
        reader = csv.reader(fh, delimiter=",")

        data: List[float] = []

        for item in reader:

            if not item:
                continue
            else:
                data.append(float((item[1])))

        return data


def copy_file_from_to(source: Path, dst: Path) -> None:
    """Copy the source image  to the dst image

    Parameters
    ----------
    source: The source image
    dst: The destination image

    Returns
    -------

    """
    shutil.copy(src=source, dst=dst)


def copy_files_from_to(sources: List[Path], dsts: List[Path]):
    for src, dist in zip(sources, dsts):
        copy_file_from_to(source=src, dst=dist)


def move_file_from_to(src: Path, dist: Path):
    shutil.move(src=src, dst=dist)


def move_files_from_to(sources: List[Path], dsts: List[Path]):
    if len(sources) != len(dsts):
        raise ValueError("Invalid size of sources and destinations")

    for src, dist in zip(sources, dsts):
        move_file_from_to(src=src, dist=dist)


def rename_file(src: Path, dst: Path):
    os.rename(src=src, dst=dst)


def compare_dirs(dir_path_1: Path, dir_path_2: Path,
                 ignore_suffixes: bool,
                 sort_contents: bool = True) -> None:
    """Compare the contents of the two directories.
    Throws a ValueError exception is the directories
    do not have the same contents according to the specified
    criteria.

    Parameters
    ----------
    dir_path_1: The first directory
    dir_path_2: The second directory
    ignore_suffixes: Flag indicating whether file suffixes should be ignored
    sort_contents: Flag indicating whether the contents of the directories should be sorted
    before the comparison

    Returns
    -------

    None
    """

    files_1 = os.listdir(dir_path_1)
    files_2 = os.listdir(dir_path_2)

    if len(files_1) != len(files_2):
        raise ValueError(f"Directories don't have the same contents. Size {len(files_1)} != {len(files_2)}")

    if sort_contents:
        files_1.sort()
        files_2.sort()

    if ignore_suffixes:

        for file_1, file_2 in zip(files_1, files_2):
            name1 = file_1.split(".")[0]
            name2 = file_2.split(".")[0]

            if str(name1) != str(name2):
                raise ValueError(f"Names not equal: {name1} != {name2}")
    else:
        for file_1, file_2 in zip(files_1, files_2):
            if str(file_1) != str(file_2):
                raise ValueError(f"Names not equal: {file_1} != {file_2}")


def get_all_files(dir_path: Path, file_formats: List[str],
                  with_batch_structure: bool = False,
                  skip_dirs: bool = False) -> List[Path]:
    """Get the image files in the given image directory that have
    the specified image format.

    Parameters
    ----------
    with_batch_structure: Flag indicating that the given path is structured
    with subdirectories
    img_dir: The image directory
    file_formats: The image formats

    Returns
    -------
    An instance of List[Path]
    """

    if with_batch_structure and skip_dirs:
        raise ValueError("Data has batch structure but skip_dirs=True")

    # load the corrosion images
    img_files = os.listdir(dir_path)

    files: List[Path] = []

    for filename in img_files:
        if os.path.isfile(dir_path / filename):

            if filename.split(".")[-1].lower() in file_formats:
                files.append(dir_path / filename)
            else:
                continue

        elif with_batch_structure:
            if os.path.isdir(dir_path / filename):
                dir_img_files = os.listdir(dir_path / filename)

                for img_filename in dir_img_files:
                    if os.path.isfile(dir_path / filename / img_filename) and \
                            img_filename.split(".")[-1].lower() in file_formats:
                        files.append(dir_path / filename / img_filename)
        elif skip_dirs:
            continue
        else:
            raise ValueError(f"The given path {dir_path / filename} is a directory but no batches are assumed")

    return files



from .image_statistics import (compute_image_statistics,
                               compute_image_channels_means,
                               compute_image_channels_variance,
                               compute_image_channels_median,
                               fit_gaussian_distribution,
                               fit_gaussian_distribution_on_image,
                               kl_divergence,
                               jensen_shannon_distance)


import numpy as np
from PIL import Image
from pathlib import Path
from typing import Callable, List, Union, TypeVar
from io import BytesIO

from navalmartin_mir_vision_utils import get_img_files

from navalmartin_mir_vision_utils.mir_vision_config import WITH_TORCH, WITH_CV2
from navalmartin_mir_vision_utils.mir_vision_types import TorchTensor
from navalmartin_mir_vision_utils.exceptions import InvalidPILImageMode, InvalidConfiguration
from navalmartin_mir_vision_utils.image_enums import (ImageFileEnumType, ImageLoadersEnumType,
                                                      IMAGE_LOADERS_TYPES_STR, IMAGE_STR_TYPES, VALID_PIL_MODES_STR)
from navalmartin_mir_vision_utils.mir_vison_io.file_utils import ERROR

if WITH_TORCH:
    import torch
    import torchvision
    from torchvision import transforms


if WITH_CV2:
    import cv2


def load_img(path: Path, transformer: Callable = None,
             loader: ImageLoadersEnumType = ImageLoadersEnumType.PIL):
    """Load the image from the given path

    Parameters
    ----------
    path: The path to the image
    transformer: Callable object that applies transformations on the image
    loader: How to read the image currently only CV2 or PIL

    Returns
    -------

    If not transform is used it returns an PIL.Image object.
    Otherwise, it returns the datasets type supported by the
    transform
    """

    if loader.name.upper() not in IMAGE_LOADERS_TYPES_STR:
        raise ValueError(f"Invalid image loader={loader.name.upper()} not in {IMAGE_LOADERS_TYPES_STR}")

    if loader == ImageLoadersEnumType.PIL:
        return load_image_as_pillow(path=path, transformer=transformer)

    if loader == ImageLoadersEnumType.PIL_NUMPY:
        return load_image_as_numpy(path=path, transformer=transformer)

    if loader == ImageLoadersEnumType.CV2:
        if WITH_CV2:
            return load_image_cv2(path=path, transformer=transformer)
        else:
            raise InvalidConfiguration(message="opencv-python was not found. Cannot import cv2")

    if loader == ImageLoadersEnumType.PYTORCH_TENSOR:
        if WITH_TORCH:
            return load_image_pytorch_tensor(path=path, transformer=transformer)
        else:
            raise InvalidConfiguration(message="PyTorch was not found.")

    return None


def load_pil_image_from_byte_string(image_byte_string: bytes,
                                    open_if_verify_success: bool = True) -> Image:
    """Loads a PIL.Image from the given byte string

    Parameters
    ----------
    image_byte_string: The byte string representing the image
    open_if_verify_success: Whether to reopen the Image after image.verify()
    is called

    Returns
    -------

    An instance of PIL.Image
    """
    try:
        image = Image.open(BytesIO(image_byte_string))
        image.verify()

        # we need to reopen after verify
        # see this:
        # https://stackoverflow.com/questions/3385561/python-pil-load-throwing-attributeerror-nonetype-object-has-no-attribute-rea
        if open_if_verify_success:
            image = Image.open(BytesIO(image_byte_string))

        return image
    except (IOError, SyntaxError) as e:
        print(f"{ERROR} the image_byte_string is corrupted")
        print(f"Exception message {str(e)}")
        return None


def load_images(path: Path, transformer: Callable = None,
                loader: ImageLoadersEnumType = ImageLoadersEnumType.PIL,
                img_formats: tuple = IMAGE_STR_TYPES) -> List:
    """Loads all the images in the specified path

    Parameters
    ----------
    path: The path to load the images from
    transformer: how to transform the images
    loader: The type of the laoder either PIL or CV2
    img_formats: The image format

    Returns
    -------
    A list of images. The actual type depends on the type of the
    """

    # get all the image files
    img_files = get_img_files(base_path=path, img_formats=img_formats)

    if len(img_files) == 0:
        raise ValueError(f"{path} does  not have images with formats {img_formats}")

    images = []
    # load every image in the Path
    for img in img_files:
        images.append(load_img(path=img,
                               transformer=transformer,
                               loader=loader))

    return images


def load_images_from_paths(imgs: List[Path], transformer: Callable,
                           loader: ImageLoadersEnumType = ImageLoadersEnumType.PIL):
    if len(imgs) == 0:
        raise ValueError("Empty images paths")

    if loader not in IMAGE_LOADERS_TYPES_STR:
        raise ValueError(f"Invalid loader. Loader={loader} not in {IMAGE_LOADERS_TYPES_STR}")

    imgs_data = []
    for img in imgs:
        imgs_data.append(load_img(path=img, transformer=transformer, loader=loader))

    return imgs_data  


def load_image_as_pillow(path: Path, transformer: Callable = None) -> Image:
    """Load the image in the specified path as Pillow.Image object

    Parameters
    ----------
    path
    transformer

    Returns
    -------

    """
    x = Image.open(path)

    if transformer is not None:
        x = transformer(x)
    return x


def load_image_as_numpy(path: Path, transformer: Callable = None) -> np.array:
    x = Image.open(path)
    if transformer is not None:
        x = transformer(x)
    return np.array(x)


def load_image_cv2(path: Path, transformer: Callable = None):
    """Load an image as OpenCV matrix. If WITH_CV2 is False
    throws InvalidConfiguration

    Parameters
    ----------
    path: Path to the image
    transformer: Transformer to apply on loading

    Returns
    -------
    OpenCV image matrix
    """

    if not WITH_CV2:
        print("opencv-python is not installed.")
        raise InvalidConfiguration(message="opencv-python is not installed.")

    image = cv2.imread(str(path))

    if transformer is not None:
        image = transformer(image)

    return image


def load_image_pytorch_tensor(path: Path, transformer: Callable = None) -> TorchTensor:
    """Load the image from the specified path.  If WITH_TORCH is False
    throws InvalidConfiguration

    Parameters
    ----------
    path
    transformer

    Returns
    -------


    """

    if not WITH_TORCH:
        print("PyTorch is not installed.")
        raise InvalidConfiguration(message="PyTorch is not installed.")

    with Image.open(path) as image:

        if transformer is None:
                transform_to_torch = transforms.Compose([transforms.ToTensor()])
                return transform_to_torch(image)

        x = image
        x = transformer(x)

        if isinstance(x, torch.Tensor):
            return x

        transform_to_torch = transforms.Compose([transforms.ToTensor()])
        return transform_to_torch(x)
   

def load_images_as_torch(x: List[Path], y_train: List[int],
                         transformer: Callable) -> tuple:
    """Load the images in the path as torch tensors

    Parameters
    ----------
    x: A list of image files
    y_train: The lebels associated with evey image
    transformer: Transform to apply when loading the images.
    Usually this will be transforms.Compose

    Returns
    -------
    A tuple of torch.Tensors
    """

    if not WITH_TORCH:
        print("load_image_pytorch_tensor is not available as PyTorch was not found.")
        raise
        return None

    data = [load_img(img_path, transformer) for img_path in x]
    return torch.stack(data), torch.tensor(y_train, dtype=torch.uint8)


from PIL import Image
import numpy as np
import io


def pil2ndarray(image: Image) -> np.ndarray:
    """Convert the Pillow image into a numpy array

    Parameters
    ----------
    image: The Pillow image to transform

    Returns
    -------

    """
    return np.asarray(image)


def pil_image_to_bytes_string(image: Image) -> bytes:
    """Returns the byte string of the provided image.
    Code adapted from https://stackoverflow.com/questions/33101935/convert-pil-image-to-byte-array
    Parameters
    ----------
    image: The PIL.Image

    Returns
    -------
    A string that represents the image bytes
    """

    img_byte_arr = io.BytesIO()
    # image.save expects a file-like as a argument
    image.save(img_byte_arr, format=image.format)
    # Turn the BytesIO object back into a bytes object
    imgByteArr = img_byte_arr.getvalue()
    return imgByteArr



"""module img_utils. Various image utilities

"""
import os
import shutil
from pathlib import Path
from io import BytesIO
from typing import Callable, List, Union, TypeVar
from PIL import Image

import numpy as np


from navalmartin_mir_vision_utils.exceptions import InvalidPILImageMode
from navalmartin_mir_vision_utils.image_enums import (ImageFileEnumType, ImageLoadersEnumType,
                                                      IMAGE_LOADERS_TYPES_STR, IMAGE_STR_TYPES, VALID_PIL_MODES_STR)
from navalmartin_mir_vision_utils.mir_vison_io.file_utils import ERROR


def is_valid_pil_image_from_bytes_string(image_byte_string: bytes, open_if_verify_succcess: bool = True) -> Image:
    """Check if the provided bytes correspond to a valid
    PIL.Image. If IOError or  SyntaxError is raised it returns None

    Parameters
    ----------
    image_byte_string: The string bytes that correspond to an image

    Returns
    -------

    An instance of a PIL.Image if the image is valid or None
    """
    try:
        image = Image.open(BytesIO(image_byte_string))
        image.verify()
        # we need to reopen after verify
        # see this:
        # https://stackoverflow.com/questions/3385561/python-pil-load-throwing-attributeerror-nonetype-object-has-no-attribute-rea
        if open_if_verify_succcess:
            image = Image.open(BytesIO(image_byte_string))
        return image
    except (IOError, SyntaxError) as e:
        print(f"{ERROR} the image_byte_string is corrupted")
        print(f"Exception message {str(e)}")
        return None


def is_valid_pil_image_file(image: Path) -> Image:
    """Check if the given image is a valid Pillow image

    Parameters
    ----------
    image: The image filename

    Returns
    -------
    An instance of a PIL.Image if the image is valid or None
    """

    try:
        img = Image.open(image)
        img.verify()
        return img
    except (IOError, SyntaxError) as e:
        print(f"{ERROR} the file {image} is corrupted")
        return None


def get_pil_image_size(image: Image) -> tuple:
    """Returns the width, height of the given Pillow image

    Parameters
    ----------
    image: The PIL.Image

    Returns
    -------
    A tuple representing width, height
    """

    if image is None:
        raise ValueError("The provided image is None")

    return image.size


def list_image_files(base_path: Path,
                     valid_exts: Union[List , tuple] = IMAGE_STR_TYPES,
                     contains: str = None) -> Path:
    """Generator that returns all the images in the given
    base_path

    Parameters
    ----------
    base_path: Base path to look for image files
    valid_exts: Extensions to use
    contains: String that the filename should contain

    Returns
    -------

    """

    if isinstance(valid_exts, tuple):
        valid_exts = list(valid_exts)

    for i, item in enumerate(valid_exts):
        if isinstance(item, ImageFileEnumType):
            valid_exts[i] = f'.{item.name.lower()}'

    if not isinstance(valid_exts, tuple):
        valid_exts = tuple(valid_exts)

    # loop over the directory structure
    for (rootDir, dirNames, filenames) in os.walk(base_path):
        # loop over the filenames in the current directory
        for filename in filenames:
            # if the contains string is not none and the filename does not contain
            # the supplied string, then ignore the file
            if contains is not None and filename.find(contains) == -1:
                continue

            # determine the file extension of the current file
            ext = filename[filename.rfind("."):].lower()

            # check to see if the file is an image and should be processed
            if valid_exts is None or ext.endswith(valid_exts):
                # construct the path to the image and yield it
                image_path = os.path.join(rootDir, filename)
                yield image_path


def get_img_files(base_path: Path,
                  img_formats: Union[List , tuple] = IMAGE_STR_TYPES) -> List[Path]:
    """Get the image files in the given image directory that have
    the specified image format.

    Parameters
    ----------
    base_path: The image directory
    img_formats: The image formats

    Returns
    -------
    An instance of List[Path]
    """

    return list(list_image_files(base_path=base_path, valid_exts=img_formats))


def remove_metadata_from_image(image: Image, new_filename: Path) -> None:
    """Remove the metadata from the given image
    and saves it to the new location

    Parameters
    ----------
    image: The image to use
    new_filename: Where to store the new image

    Returns
    -------

    """

    data = list(image.getdata())
    image_without_exif = Image.new(image.mode, image.size)
    image_without_exif.putdata(data)
    image_without_exif.save(new_filename)


def delete_image_if(image: Image, size: tuple, direction: str) -> None:
    """Delete the image specified in the path that its file does not
    satisfy the direction. Direction can be any of the following:
    - >_both: Both coordinates are greater than
    - >_x: x coordinate is greater than
    - >_y: y coordinate is greater than
    - >_any: Any of the coordinates is less than
    - <_both: Both coordinates are less than
    - <_x: x coordinate is less than
    - <_y: y coordinate is less than
    - <_any: Any of the coordinates is less than


    Parameters
    ----------
    direction
    image
    size

    Returns
    -------

    """

    img_size_width = image.width
    img_size_height = image.height

    if direction == ">_any" or direction == "<_any":

        if img_size_width < size[0] or \
                img_size_width > size[0] or \
                img_size_height < size[1] or \
                img_size_width > size[1]:
            os.remove(path=image.filename)
    else:
        raise ValueError(f"Direction {direction} not implemented")


def copy_img_from_to(source: Path, dst: Path) -> None:
    """Copy the source image  to the dst image

    Parameters
    ----------
    source: The source image
    dst: The destination image

    Returns
    -------

    """
    shutil.copy(src=source, dst=dst)


def save_img(image: Image, filename: Path, img_format: str = None) -> None:
    """Save the given image on the given path

    Parameters
    ----------
    image: The image to save
    filename: The name of the image file
    img_format: The format of the image
    Returns
    -------

    """

    if format is None or format == "":
        image.save(filename)
    else:
        image.save(filename, format=img_format)


def show_pil_image(image: Image) -> None:
    """Show the image depending on the
    type

    Parameters
    ----------
    image
    img: The image to show

    Returns
    -------

    """

    if image is None:
        raise ValueError("The provided image is None")

    image.show()


def save_img_from_str(img_str: str, encoding: str,
                      img_format: str, path: Path,
                      img_height: int, img_width: int,
                      mode: str = 'RGB') -> None:
    """

    Parameters
    ----------
    img_str: The decoded image string
    encoding: The encoding used to decode the image to a string
    img_format: The format of the image
    path: The path to store the image
    img_height: The height of the image
    img_width: The width of the image
    mode: The mode of an image is a string which defines the
    type and depth of a pixel in the image.
    Each pixel uses the full range of the bit depth.
    So a 1-bit pixel has a range of 0-1, an 8-bit pixel has a range of 0-255 and so on.
    The current release supports the following standard modes:
    For more information check https://pillow.readthedocs.io/en/stable/handbook/concepts.html#concept-modes

    Returns
    -------

    None
    """

    # convert bytes data to PIL Image object

    if mode not in VALID_PIL_MODES_STR:
        raise InvalidPILImageMode(mode=mode)

    if encoding is None:
        img_bytes = str.encode(img_str)
    else:
        img_bytes = img_str.encode(encoding)

    img = Image.frombytes(mode, (img_width, img_height), img_bytes, 'raw')

    img.save(fp=path, format=img_format)



"""module asymmetric_generalized_gaussian
Implements asymmetric generalized gaussian distribution.
The initial implementation is taken from
 https://github.com/ocampor/image-quality

"""
from enum import Enum
from typing import Callable

import numpy
import scipy.optimize
import scipy.special


class DistributionSide(Enum):
    right = 1
    left = 2


def find_root(fun: Callable, x0: float = 0.2) -> scipy.optimize.OptimizeResult:
    x0 = numpy.array([x0])
    return scipy.optimize.root(fun, x0)


class AsymmetricGeneralizedGaussian:
    def __init__(self, x: numpy.ndarray):
        assert isinstance(x, numpy.ndarray)

        self.x = x
        self._alpha = None

    def _x(self, side: DistributionSide):
        if side == DistributionSide.left:
            idx = numpy.where(self.x < 0)
        elif side == DistributionSide.right:
            idx = numpy.where(self.x >= 0)
        else:
            raise ValueError(f"Side {side} was not recognized")

        return self.x[idx]

    @property
    def x_left(self):
        return self._x(DistributionSide.left)

    @property
    def x_right(self):
        return self._x(DistributionSide.right)

    def _sigma(self, side: DistributionSide):
        if side == DistributionSide.right:
            _x = self.x_right
        elif side == DistributionSide.left:
            _x = self.x_left
        else:
            raise ValueError(f"Side {side} was not recognized")

        return numpy.sqrt(self.mean_squares(_x))

    @property
    def sigma_left(self):
        return self._sigma(DistributionSide.left)

    @property
    def sigma_right(self):
        return self._sigma(DistributionSide.right)

    @property
    def gamma(self):
        return self.sigma_left / self.sigma_right

    @property
    def r_hat(self):
        return numpy.abs(self.x).mean() ** 2 / self.mean_squares(self.x)

    @property
    def R_hat(self):
        return (
            self.r_hat
            * (self.gamma ** 3 + 1)
            * (self.gamma + 1)
            / (self.gamma ** 2 + 1) ** 2
        )

    @property
    def constant(self):
        return numpy.sqrt(
            scipy.special.gamma(1 / self.alpha) / scipy.special.gamma(3 / self.alpha)
        )

    @property
    def mean(self):
        return (
            (self.sigma_right - self.sigma_left)
            * self.constant
            * (
                scipy.special.gamma(2 / self.alpha)
                / scipy.special.gamma(1 / self.alpha)
            )
        )

    @property
    def alpha(self):
        if self._alpha is None:
            raise NotImplementedError(
                "The distribution has no alpha estimated. Run method fit() to calculate."
            )
        return self._alpha

    @staticmethod
    def mean_squares(x: numpy.array):
        return numpy.square(x).mean()

    @staticmethod
    def phi(alpha):
        return scipy.special.gamma(2 / alpha) ** 2 / (
            scipy.special.gamma(1 / alpha) * scipy.special.gamma(3 / alpha)
        )

    def estimate_alpha(self, x0: float = 0.2) -> float:
        try:
            solution = find_root(lambda alpha: self.phi(alpha) - self.R_hat, x0)
            assert solution.success
            return solution.x.item()
        except ValueError:
            raise ValueError(
                f"More than one solution was found for phi(alpha) - {self.R_hat}"
            )

    def fit(self, x0: float = 0.2) -> "AsymmetricGeneralizedGaussian":
        self._alpha = self.estimate_alpha(x0)
        return self



import unittest
from pathlib import Path
from src.navalmartin_mir_vision_utils.image_utils import get_img_files
from src.navalmartin_mir_vision_utils.image_enums import ImageFileEnumType


class TestImageUtils(unittest.TestCase):

    def test_get_img_files(self):
        files = get_img_files(img_dir=Path('/imutils/test_data'),
                              img_formats=(ImageFileEnumType.PNG,
                                           ImageFileEnumType.JPG))

        self.assertEqual(2, len(files))


if __name__ == '__main__':
    unittest.main()



"""
Various utilities for working with images in the _mir_ project.
"""
__version__ = "0.0.12"


from .image_utils import (is_valid_pil_image_from_bytes_string,
                          is_valid_pil_image_file,
                          show_pil_image,
                          get_img_files,
                          get_pil_image_size)
from .image_loaders import load_img
from .image_transformers import pil2ndarray, pil_image_to_bytes_string
from .image_enums import (ImageFileEnumType,
                          IMAGE_STR_TYPES,
                          ImageLoadersEnumType,
                          IMAGE_LOADERS_TYPES_STR,
                          ValidPillowEnumType,
                          VALID_PIL_MODES_STR)

from .exceptions import InvalidConfiguration







WITH_TORCH = False
WITH_CV2 = False

try:
    import torch
    import torchvision
    from torchvision import transforms

    WITH_TORCH = True
except ModuleNotFoundError as e:
    print(
        f"WARNING: An exception was raised whilst importing torch, torchvision. Message {str(e)}")
    print("WARNING: mir-vision-utils will not use torch and torchvision")
    pass

try:
    import cv2
    WITH_CV2 = True
except ModuleNotFoundError as e:
    print(
        f"WARNING: An exception was raised whilst importing cv2 for OpenCV. Message {str(e)}")
    print("WARNING: mir-vision-utils will not use OpenCV")
    pass


from enum import Enum


class ImageFileEnumType(Enum):

    INVALID = 0
    JPG = 1
    JPEG = 2
    PNG = 3
    BMP = 4
    TIF = 5
    TIFF = 6


IMAGE_STR_TYPES = (".jpg", ".jpeg", ".png", ".bmp", ".tif", ".tiff")


class ImageLoadersEnumType(Enum):

    INVALID = 0
    PIL = 1
    CV2 = 2
    PIL_NUMPY = 3
    PYTORCH_TENSOR = 4


IMAGE_LOADERS_TYPES_STR = ('PIL', 'CV2', 'PIL_NUMPY', 'PYTORCH_TENSOR')


class ValidPillowEnumType(Enum):

    INVALID = 0
    ONE = 1
    CMYK = 2
    F = 3
    HSV = 4
    I = 5
    L = 6
    LAB = 7
    P = 8
    RGB = 9
    RGBA = 10
    RGBX = 11
    YCbCr = 12


VALID_PIL_MODES_STR = ("1", "CMYK", "F", "HSV", "I", "L", "LAB", "P", "RGB", "RGBA", "RGBX", "YCbCr")


"""module brisque. Implementation of the
BRISQUE algorithm. Original implementation is
taken from: https://github.com/ocampor/image-quality

"""
import os
import pickle
import typing
import warnings
from enum import Enum

import PIL.Image
import numpy
import scipy.signal
import skimage.color
import skimage.transform
from libsvm import svmutil

from navalmartin_mir_vision_utils.statistics.distributions.asymmetric_generalized_gaussian import AsymmetricGeneralizedGaussian
from navalmartin_mir_vision_utils.statistics.statistics_utils import gaussian_kernel2d
from navalmartin_mir_vision_utils.image_transformers import pil2ndarray
from navalmartin_mir_vision_utils.image_quality.models import MODELS_PATH

VALID_BRISQUE_IMAGE_FORMATS = ['JPG', 'JPEG']

with open(os.path.join(MODELS_PATH, "normalize.pickle"), "rb") as file:
    scale_parameters = pickle.load(file)

model = svmutil.svm_load_model(os.path.join(MODELS_PATH, "brisque_svm.txt"))


class MscnType(Enum):
    mscn = 1
    horizontal = 2
    vertical = 3
    main_diagonal = 4
    secondary_diagonal = 5


class Brisque:
    _local_mean = None
    _local_deviation = None
    _mscn = None
    _features = None

    def __init__(
        self,
        image: typing.Union[PIL.Image.Image, numpy.ndarray],
        kernel_size: int = 7,
        sigma: float = 7 / 6,
    ):
        self.image = pil2ndarray(image)

        if self.image.shape[-1] == 3:
            self.image = skimage.color.rgb2gray(self.image)

        self.kernel_size = kernel_size
        self.sigma = sigma
        self.kernel = gaussian_kernel2d(kernel_size, sigma)

    @property
    def local_mean(self):
        if self._local_mean is None:
            self._local_mean = scipy.signal.convolve2d(self.image, self.kernel, "same")
        return self._local_mean

    @property
    def local_deviation(self):
        if self._local_deviation is None:
            sigma = numpy.square(self.image)
            sigma = scipy.signal.convolve2d(sigma, self.kernel, "same")
            self._local_deviation = numpy.sqrt(
                numpy.abs(numpy.square(self.local_mean) - sigma)
            )
        return self._local_deviation

    @property
    def mscn(self):
        if self._mscn is None:
            c = 1 / 255
            self._mscn = (self.image - self.local_mean) / (self.local_deviation + c)
        return self._mscn

    @property
    def mscn_horizontal(self):
        return self.mscn[:, :-1] * self.mscn[:, 1:]

    @property
    def mscn_vertical(self):
        return self.mscn[:-1, :] * self.mscn[1:, :]

    @property
    def mscn_diagonal(self):
        return self.mscn[:-1, :-1] * self.mscn[1:, 1:]

    @property
    def mscn_secondary_diagonal(self):
        return self.mscn[1:, :-1] * self.mscn[:-1, 1:]

    @property
    def features(self):
        return numpy.concatenate(
            [self.calculate_features(mscn_type) for mscn_type in MscnType]
        )

    def get_coefficients(self, mscn_type: MscnType):
        coefficients = {
            MscnType.mscn: self.mscn,
            MscnType.horizontal: self.mscn_horizontal,
            MscnType.vertical: self.mscn_vertical,
            MscnType.main_diagonal: self.mscn_diagonal,
            MscnType.secondary_diagonal: self.mscn_secondary_diagonal,
        }
        return coefficients[mscn_type]

    def calculate_features(self, mscn_type: MscnType):
        agg = AsymmetricGeneralizedGaussian(self.get_coefficients(mscn_type)).fit()
        if mscn_type == MscnType.mscn:
            var = numpy.mean(
                [numpy.square(agg.sigma_left), numpy.square(agg.sigma_right)]
            )
            return numpy.array([agg.alpha, var])

        return numpy.array(
            [
                agg.alpha,
                agg.mean,
                numpy.square(agg.sigma_left),
                numpy.square(agg.sigma_right),
            ]
        )


def scale_features(features: numpy.ndarray) -> numpy.ndarray:
    _min = numpy.array(scale_parameters["min_"])
    _max = numpy.array(scale_parameters["max_"])
    return -1 + (2.0 / (_max - _min) * (features - _min))


def calculate_features(image: PIL.Image, kernel_size, sigma) -> numpy.ndarray:
    brisque = Brisque(image, kernel_size=kernel_size, sigma=sigma)
    # WARNING: The algorithm is very sensitive to rescale
    # FIXME: this is empirically the best configuration; however, scikit-image warns about bi-quadratic implementation.
    #    Fix this warning error in version of scikit-image 0.16.0.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        downscaled_image = skimage.transform.rescale(
            brisque.image,
            1 / 2,
            order=2,
            mode="constant",
            anti_aliasing=False,
            multichannel=False,
        )
    downscaled_brisque = Brisque(downscaled_image, kernel_size=kernel_size, sigma=sigma)
    features = numpy.concatenate([brisque.features, downscaled_brisque.features])
    scaled_features = scale_features(features)
    return scaled_features


def predict(features: numpy.ndarray) -> float:
    x, idx = svmutil.gen_svm_nodearray(
        features, isKernel=(model.param.kernel_type == svmutil.PRECOMPUTED)
    )
    nr_classifier = 1
    prob_estimates = (svmutil.c_double * nr_classifier)()
    return svmutil.libsvm.svm_predict_probability(model, x, prob_estimates)


def score(image: PIL.Image.Image, kernel_size=7, sigma=7 / 6) -> float:
    """Calculate the BRISQUE score for the given image. Currently, only
    JPG images are supported

    Parameters
    ----------
    image: The Pillow image to calculate its quality
    kernel_size
    sigma

    Returns
    -------

    The BRISQUE score
    """

    if image is None:
        raise ValueError("Pillow Image is None")

    if image.format not in VALID_BRISQUE_IMAGE_FORMATS:
        raise ValueError(f"Current image format {image.format} is not in {VALID_BRISQUE_IMAGE_FORMATS}")

    scaled_features = calculate_features(image, kernel_size, sigma)
    return predict(scaled_features)



import unittest

from pathlib import Path

from src.navalmartin_mir_vision_utils.statistics.image_statistics import compute_image_channels_means
from src.navalmartin_mir_vision_utils.statistics.image_statistics import compute_image_channels_variance
from src.navalmartin_mir_vision_utils.image_utils import load_img
from src.navalmartin_mir_vision_utils.image_enums import ImageLoadersEnumType


class TestImageStatistics(unittest.TestCase):

    def test_compute_image_channels_means(self):

        image = load_img(path=Path('/imutils/test_data/corrosion_4.png'),
                         loader=ImageLoadersEnumType.PIL)
        means = compute_image_channels_means(image=image)
        self.assertEqual(len(means), 4)

    def test_compute_image_channels_variance(self):

        image = load_img(path=Path('/imutils/test_data/corrosion_4.png'),
                         loader=ImageLoadersEnumType.PIL)
        means = compute_image_channels_variance(image=image)
        self.assertEqual(len(means), 4)


if __name__ == '__main__':
    unittest.main()


"""module statistics_utils Various utilities
used around the library. The first implementation is
taken from: https://github.com/ocampor/image-quality

"""
import numpy as np


def normalize_kernel(kernel: np.ndarray) -> np.ndarray:
    """Normalize the given kernel

    Parameters
    ----------
    kernel

    Returns
    -------

    """
    return kernel / np.sum(kernel)


def gaussian_kernel2d(kernel_size, sigma: float):
    y, x = np.indices((kernel_size, kernel_size)) - int(kernel_size / 2)
    kernel = (
        1
        / (2 * np.pi * sigma ** 2)
        * np.exp(-(x ** 2 + y ** 2) / (2 * sigma ** 2))
    )
    return normalize_kernel(kernel)
